#include <Windows.h>
#include <sys\timeb.h>
#include <SDL.h>
#include <stdio.h>
#include <winnt.h>

#include "Constants.h"
#include "Timing.h"
#include "Io.h"
#include "ULABus.h"
#include "Memory.h"
#include "Cpu.h"
#include "Log.h"
#include "Sound.h"
#include "TapUi.h"
#include "PokeUi.h"
#include "Notification.h"
#include "SaveStates.h"
#include "Video.h"


static SDL_Surface* _surface;
static SDL_Renderer* _renderer;
static SDL_Window* _window;

static Uint32 _videoSurfacePixelWidth;
static Uint32 _videoSurfaceWidthZoomedBytes;

static BOOL _videoIsInitialized = FALSE;
SDL_Rect _entireWindowRectangle;

static Uint32 _rendererFlags = SDL_RENDERER_SOFTWARE;
enum frameRenderMode { ForceSkipPixels, DrawPixels };

HANDLE _callbackTimerHandle;
HANDLE _videoFrameMutex;

// we free all pixel cache pointers upon shutdown, and sometimes
// a frame is still drawing
// this mutex thus isolates screen drawing from the pointers freeing
HANDLE _pixelCacheMutex;

static volatile char _isDestroying;
enum videoMode _videoMode;
enum videoSoundSwitch _videoSoundSwitch = VideoSoundSwitch_OnLag;

static volatile char _timerCallbackDestroyed;
static SDL_TimerID _timer;

static double _lagAllowanceBeforeFrameskip = 1.03f;

struct timeb _videoStart, _videoEnd;
struct timeb _currentFrameStart;

static Uint64 _framesAtNormalSpeedSinceNormalSpeed = 0;
static Uint64 _framesSkipped = 0;
static Uint64 _framesWithNoVideoChanges = 0;
struct timeb _videoStartSinceNormalSpeed;
static Uint8 _lastFrameWasNormalSpeed;

static Uint8 _frameskipIsEnabled = 1;
static Uint64 _consecutiveSkippedFrames = 0;

static Uint64 _frames = 0;
static Uint64 _framesAtNormalSpeed = 0;
static Uint64 _tstates = 0;
static Uint8 _zoom = 1;

static Uint8 _isFlashInverse;
static Uint8 _wasFlashInverse;

// stores one border colour per scan line while CPU is being run
// this is so that during rendering, each scan line can read its border colour from here
//     without this, each video frame will show a solid border of the last border colour
//     set during current frame
//     that would be bad because tape loading lines would not show, for example
#define BORDER_SCANLINES (SPECTRUM_SCREEN_HEIGHT)
static Uint8 _borderColourPerScanline[BORDER_SCANLINES];
static Uint8 _videoIsBorderSolid = 0;

static SDL_Rect _topBorder;
static SDL_Rect _bottomBorder;
static SDL_Rect _leftBorder;
static SDL_Rect _rightBorder;


// this is used when CPU is fast, and we MUST show one video frame out of N video frames
// despite the fact that no video changes may have taken place
static Uint8 _forceFrameRender = 0;

#define _DEBUG_BUFFER_SIZE 256
static char _debugBuffer[_DEBUG_BUFFER_SIZE];

static volatile Uint8 _stateSaveRequested = 0;
static volatile Uint8 _stateLoadRequested = 0;
static volatile Uint8 _stateSaveSlot = 0;
static volatile Uint8 _stateLoadSlot = 0;

static Uint8 _palette[16*3] = {
    0x00, 0x00, 0x00,   // black
    0x00, 0x22, 0xc7,   // blue
    0xd6, 0x28, 0x16,   // red
    0xd4, 0x33, 0xc7,   // magenta
    0x00, 0xc5, 0x25,   // green
    0x00, 0xc7, 0xc9,   // cyan
    0xcc, 0xc8, 0x2a,   // yellow
    0xca, 0xca, 0xca,   // white

    // bright versions
    0x00, 0x00, 0x00,   // black
    0x00, 0x2b, 0xfb,   // blue
    0xff, 0x33, 0x1c,   // red
    0xff, 0x40, 0xfc,   // magenta
    0x00, 0xf9, 0x2f,   // green
    0x00, 0xfb, 0xfe,   // cyan
    0xff, 0xfc, 0x36,   // yellow
    0xff, 0xff, 0xff,   // white
};

// stores pre-computed values so we only convert RGB -> RGBA once
static Uint32 _rgbaPalette[16];

// stores pre-computed pointers to screen addresses of the start of each ZX Spectrum paper scan line
static Uint32* _paperScanlineStartPointers[VISIBLE_SPECTRUM_SCREEN_HEIGHT];

// stores pre-computed pointers to screen addresses of the start of each ZX Spectrum visible screen
static Uint32* _visibleScreenScanlineStartPointers[VISIBLE_SPECTRUM_SCREEN_HEIGHT];
// stores pre-computed pointers to screen addresses of the start of each ZX Spectrum border to the right of paper
static Uint32* _visibleScreenScanlineRightBorderStartPointers[VISIBLE_SPECTRUM_SCREEN_HEIGHT];

// stores pre-computed addresses of attribute entries for each 8x8 pixel ZX Spectrum square
static int _attributeAreaAddresses[SPECTRUM_PAPER_HEIGHT][SPECTRUM_PAPER_WIDTH/8];

// stores pre-rendered zoomed ZX Spectrum screen data bytes, according to pixel data, ink colour, paper colour, inverse status
// format:
//     [screen data byte] [ink] [paper] [inverse] [bright] --> RGBA data buffer
// example:
//     assume zoom=2
//     [00110011b] [1] [2] [0] [0]  -->   
//         paperRGBA, paperRGBA, paperRGBA, paperRGBA, inkRGBA, inkRGBA, inkRGBA, inkRGBA, paperRGBA, paperRGBA, paperRGBA, paperRGBA, inkRGBA, inkRGBA, inkRGBA, inkRGBA
static Uint32* _zoomedRGBAByteRenderedPixels[256][SPECTRUM_COLOUR_COUNT][SPECTRUM_COLOUR_COUNT][2][2];
static Uint64 _prerenderedDataByteCount = 0;
static Uint8 _pixelCacheIsActive = 1;
static Uint8 _warnedPixelCacheIsNotActive = 0;
// holds all zeroes to serve as the stand-in if we're still rendering after the pixel cache has been cleared
static Uint32 _pixelCacheInactiveBuffer[8 * MAX_ZOOM];

static Uint32 _zoomedPaperSizeRGBA;
static Uint32 _zoomedVisibleSpectrumScanLineSizeRGBA;


void _video_log_renderer_choice() {
    if (_rendererFlags == SDL_RENDERER_SOFTWARE) {
        sprintf_s(_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Using video renderer: software");
    }
    else if (_rendererFlags == (SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC)) {
        sprintf_s(_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Using video renderer: accelerated/vsync");
    }
    else {
        sprintf_s(_debugBuffer, _DEBUG_BUFFER_SIZE - 10, "Using video renderer: UNKNOWN");
    }
    log_write(_debugBuffer);
}

double _video_get_frameskip_percentage() {
    double percentage = ((double)_framesSkipped / (double)_framesAtNormalSpeed) * 100.0f;
    return percentage;
}

double _video_get_framesrepeated_percentage() {
    double percentage = ((double)_framesWithNoVideoChanges / (double)_framesAtNormalSpeed) * 100.0f;
    return percentage;
}

Uint64 _video_get_runtime_ms() {
    struct timeb now;
    ftime(&now);

    Uint64 runTimeMs = (Uint64)(1000.0 * (now.time - _videoStart.time)
        + (now.millitm - _videoStart.millitm)) - (Uint32)VIDEO_DELAY_TIMER_START_MS;
    return runTimeMs;
}

double _video_get_actual_average_milliseconds_per_frame() {
    Uint64 timeSpentWhileCPUFasterMs = timing_get_total_real_time_spent_while_faster();
    Uint64 totalTimeAllMs = _video_get_runtime_ms();

    Uint64 idleTimeAllMs = timing_get_total_real_idle_time();
    Uint64 idleTimeWhileFasterMs = timing_get_total_real_idle_time_spent_while_faster();
    Uint64 idleTimeMs = idleTimeAllMs - idleTimeWhileFasterMs;

    Uint64 runTimeMs = totalTimeAllMs - timeSpentWhileCPUFasterMs - idleTimeAllMs;
    double result = (double)runTimeMs / (double)_framesAtNormalSpeed;
    return result;
}

void _video_handle_flash() {
    _wasFlashInverse = _isFlashInverse;
    _isFlashInverse = _frames & 0x10;

    if (_wasFlashInverse != _isFlashInverse) {
        _forceFrameRender = 1;
    }
}

// see declaration of _rgbaPalette for how this works
//
void _video_compute_rgba_palette(SDL_Surface* surface) {
    for (int i = 0; i < 16; i++) {
        Uint8 red = _palette[i * 3 + 0];
        Uint8 green = _palette[i * 3 + 1];
        Uint8 blue = _palette[i * 3 + 2];
        
        Uint32 rgba = SDL_MapRGBA(surface->format, red, green, blue, 255);
        _rgbaPalette[i] = rgba;
    }
}

// see declaration of _paperScanlineStartPointers for how this works
//
void _video_compute_scanline_pointers(SDL_Surface* surface) {
    for (int y = HIDE_TOP_BORDER_COUNT; y < VISIBLE_SPECTRUM_SCREEN_HEIGHT + HIDE_TOP_BORDER_COUNT; y++) {
        int offset = ((y - HIDE_TOP_BORDER_COUNT) * _zoom) * _videoSurfacePixelWidth + ((SPECTRUM_BORDER_WIDTH - HIDE_LEFT_BORDER_COUNT) * _zoom);
        Uint32* pixelPtr = (Uint32*)surface->pixels + offset;
        _paperScanlineStartPointers[y - HIDE_TOP_BORDER_COUNT] = pixelPtr;
    }
}

// see declaration of _visibleScreenScanlineStartPointers for how this works
// see declaration of _visibleScreenScanlineRightBorderStartPointers for how this works
//
void _video_compute_visible_screen_scanline_pointers(SDL_Surface* surface) {
    for (int y = HIDE_TOP_BORDER_COUNT; y < VISIBLE_SPECTRUM_SCREEN_HEIGHT + HIDE_TOP_BORDER_COUNT; y++) {
        int offset = ((y - HIDE_TOP_BORDER_COUNT) * _zoom) * _videoSurfacePixelWidth;
        Uint32* pixelPtr = (Uint32*)_surface->pixels + offset;

        _visibleScreenScanlineStartPointers[y - HIDE_TOP_BORDER_COUNT] = pixelPtr;
        _visibleScreenScanlineRightBorderStartPointers[y - HIDE_TOP_BORDER_COUNT] = 
            pixelPtr + (VISIBLE_SPECTRUM_LEFT_BORDER_WIDTH + SPECTRUM_PAPER_WIDTH) * _zoom;
    }
}


// see declaration of _attributeAreaAddresses for how this works
//
void _video_compute_attribute_addresses(SDL_Surface* surface) {
    for (int y = 0; y < SPECTRUM_PAPER_HEIGHT; y++) {
        for (int i = 0; i < SPECTRUM_PAPER_WIDTH; i += 8) {
            int attributeRowNumber = y / 8;
            int attributeRowOffset = attributeRowNumber * (SPECTRUM_PAPER_WIDTH / 8);
            int attributeByteOffset = attributeRowOffset + i / 8;
            int attributeAddress = SPECTRUM_VIDEO_RAM_ATTRIBUTE_AREA_START + attributeByteOffset;

            _attributeAreaAddresses[y][i / 8] = attributeAddress;
        }
    }
}

// see declaration of _zoomedRGBAByteRenderedPixels for how this works
// returns a rendered data byte (8 pixels wide ZX Spectrum line)
//
// caches the rendered data byte for future use
//
// consumer does NOT own the return pointer and MUST NOT free it
Uint32* _video_get_rendered_data_byte(Uint8 pixelByte, Uint8 inkIndex, Uint8 paperIndex, Uint8 isInverse, Uint8 isBright) {
    Uint32* rgbaBuffer = _zoomedRGBAByteRenderedPixels[pixelByte][inkIndex][paperIndex][isInverse][isBright];
    if (rgbaBuffer != NULL) {
        // we've rendered before this combination of pixel data, ink, paper, inverse, bright
        // so we can just return what we already have
        return rgbaBuffer;
    }

    Uint8 checkedZoom = _zoom;
    if (checkedZoom == 0) {
        // built-in Visual Studio static code analysis complains that the allocated 
        // buffers may be of size 0 if _zoom is 0
        checkedZoom = 1;
    }

    if (!_pixelCacheIsActive) {
        if (!_warnedPixelCacheIsNotActive) {
            _warnedPixelCacheIsNotActive = 1;
            log_write("Warning: pixel cache was access after it became inactive; returning hardcoded buffer");
        }
        // we've freed the cache memory because we're shutting down
        // so now we just return pointers to a hardcover 
        return &_pixelCacheInactiveBuffer[0];
    }

    // we have to make a new buffer
    rgbaBuffer = (Uint32*)malloc(8 * checkedZoom * sizeof(Uint32));
    if (rgbaBuffer == NULL) {
        log_write("Error: could not allocate pre-rendered data byte");
        return NULL;
    }

    Uint32 inkRGBA = _rgbaPalette[inkIndex];
    Uint32 paperRGBA = _rgbaPalette[paperIndex];

    Uint32 inkRGBABright = _rgbaPalette[inkIndex + SPECTRUM_COLOUR_COUNT];
    Uint32 paperRGBABright = _rgbaPalette[paperIndex + SPECTRUM_COLOUR_COUNT];

    Uint16 pixelByteCopy = pixelByte;
    // pre-render an 8-pixel row of ZX Spectrum bytes to a (8*zoom)-wide RGBA buffer
    for (int j = 0; j < 8; j++) {
        // j iterates through each pixel of the 8-pixel row

        for (int zoomCur = 0; zoomCur < checkedZoom; zoomCur++) {
            // write as many RGBA entries as zoom into the pre-rendered buffer
            if ((pixelByteCopy & 0x80)) {
                // pixel is set
                if (!isInverse) {
                    // it's not inverse
                    if (!isBright) {
                        rgbaBuffer[j * checkedZoom + zoomCur] = inkRGBA;
                    }
                    else {
                        rgbaBuffer[j * checkedZoom + zoomCur] = inkRGBABright;
                    }
                }
                else {
                    // it's inverse
                    if (!isBright) {
                        rgbaBuffer[j * checkedZoom + zoomCur] = paperRGBA;
                    }
                    else {
                        rgbaBuffer[j * checkedZoom + zoomCur] = paperRGBABright;
                    }
                }
            }
            else {
                // pixel is reset
                if (!isInverse) {
                    // it's not inverse
                    if (!isBright) {
                        rgbaBuffer[j * checkedZoom + zoomCur] = paperRGBA;
                    }
                    else {
                        rgbaBuffer[j * checkedZoom + zoomCur] = paperRGBABright;
                    }
                }
                else {
                    // it's inverse
                    if (!isBright) {
                        rgbaBuffer[j * checkedZoom + zoomCur] = inkRGBA;
                    }
                    else {
                        rgbaBuffer[j * checkedZoom + zoomCur] = inkRGBABright;
                    }
                }
            }
        }

        // shift next pixel into the left-most position
        pixelByteCopy = pixelByteCopy << 1;
    }

    _zoomedRGBAByteRenderedPixels[pixelByte][inkIndex][paperIndex][isInverse][isBright] = rgbaBuffer;
    _prerenderedDataByteCount++;
    return rgbaBuffer;
}

void _video_destroy_prerendered_data_bytes() {
    for (Uint16 pixelByte = 0; pixelByte < 256; pixelByte++) {
        for (Uint8 inkIndex = 0; inkIndex < SPECTRUM_COLOUR_COUNT; inkIndex++) {
            for (Uint8 paperIndex = 0; paperIndex < SPECTRUM_COLOUR_COUNT; paperIndex++) {

                for (int i = 0; i < 2; i++) {
                    for (int j = 0; j < 2; j++) {
                        Uint32** ptr = &_zoomedRGBAByteRenderedPixels[pixelByte][inkIndex][paperIndex][i][j];
                        if (*ptr != NULL) {
                            free(*ptr);
                            *ptr = NULL;
                        }
                    }
                }
            }
        }
    }
    _pixelCacheIsActive = 0;
}

void _video_destroy_prerendered_data_bytes__wrapper() {
    DWORD dwWaitResult = WaitForSingleObject(_pixelCacheMutex, INFINITE);

    switch (dwWaitResult) {
    case WAIT_OBJECT_0:
        // this thread got ownership of the mutex
        __try {
            _video_destroy_prerendered_data_bytes();
        }

        __finally {
            if (!ReleaseMutex(_pixelCacheMutex))
            {
                // an error happened
            }
        }
        break;

    case WAIT_ABANDONED:
        // this thread got ownership of an abandoned mutex
        return;
    }
}

BOOL _video_is_lagging() {
    if (_framesAtNormalSpeedSinceNormalSpeed == 0) {
        // assume first frame is not lagging and avoid divide by 0
        return FALSE;
    }

    struct timeb now;
    ftime(&now);
    Uint64 runTimeMs = (Uint64)(1000.0 * (now.time - _videoStartSinceNormalSpeed.time)
        + (now.millitm - _videoStartSinceNormalSpeed.millitm));

    double frames = (double)_framesAtNormalSpeedSinceNormalSpeed;
    double millisPerFrame = (double)runTimeMs / frames;
    double targetMillisPerFrame = (double)timing_get_target_milliseconds_per_video_frame();

    BOOL isLagging = millisPerFrame > ( targetMillisPerFrame * _lagAllowanceBeforeFrameskip);
    return isLagging;
}

void video_set_lag_allowance_before_frameskip(Uint8 percent) {
    _lagAllowanceBeforeFrameskip = 1.0f + (double)percent / 100.0f;
}

void _video_run_cpu_for_scanline() {
    Uint32 allocation = timing_get_states_per_scanline();
    Uint64 tstatesConsumed = cpu_run_many(allocation);

    if (!timing_is_faster()) {
        // we only care about statistics when we're running at normal speed
        _tstates += tstatesConsumed;

        // same for more accurate timing
        timing_report_actual_states_last_scanline((Uint32)tstatesConsumed);
    }
}

void _video_write_left_border_line(int y) {
    Uint8 paletteEntryColour = _borderColourPerScanline[y];
    Uint32 rgba = _rgbaPalette[paletteEntryColour];

    Uint32* pixelPtr = _visibleScreenScanlineStartPointers[y - HIDE_TOP_BORDER_COUNT];
    // save it for later
    Uint32* startOfScanlinePtr = pixelPtr;

    // one screen line
    SDL_memset4(pixelPtr, rgba, VISIBLE_SPECTRUM_LEFT_BORDER_WIDTH * _zoom);
}

void _video_write_right_border_line(int y) {
    Uint8 paletteEntryColour = _borderColourPerScanline[y];
    Uint32 rgba = _rgbaPalette[paletteEntryColour];

    Uint32* pixelPtr = _visibleScreenScanlineRightBorderStartPointers[y - HIDE_TOP_BORDER_COUNT];

    // save it for later
    Uint32* startOfScanlinePtr = pixelPtr;

    // one screen line
    SDL_memset4(pixelPtr, rgba, VISIBLE_SPECTRUM_RIGHT_BORDER_WIDTH * _zoom);
}

void _video_write_border_scan_line(int y) {
    Uint8 paletteEntryColour = _borderColourPerScanline[y];
    Uint32 rgba = _rgbaPalette[paletteEntryColour];

    Uint32* pixelPtr = _visibleScreenScanlineStartPointers[y - HIDE_TOP_BORDER_COUNT];
    // save it for later
    Uint32* startOfScanlinePtr = pixelPtr;

    // one screen line
    SDL_memset4(pixelPtr, rgba, VISIBLE_SPECTRUM_SCREEN_WIDTH * _zoom);

    // copy the first screen line as many times as zoom
    pixelPtr = startOfScanlinePtr;
    for (int dup = 1; dup < _zoom; dup++) {
        // move a line down on screen
        pixelPtr += _videoSurfacePixelWidth;
        memcpy_s(
            pixelPtr,                                                     // dest
            VISIBLE_SPECTRUM_SCREEN_WIDTH * _zoom * sizeof(Uint32),       // dest size
            startOfScanlinePtr,                                           // source
            VISIBLE_SPECTRUM_SCREEN_WIDTH * _zoom * sizeof(Uint32)        // source size
        );
    }
}

void _video_copy_visible_screen_line_for_zoom(int y) {
    Uint32* pixelPtr = _visibleScreenScanlineStartPointers[y - HIDE_TOP_BORDER_COUNT];
    Uint32* startOfScanlinePtr = pixelPtr;
    for (int dup = 1; dup < _zoom; dup++) {
        // move a line down on screen
        pixelPtr += _videoSurfacePixelWidth;
        memcpy_s(
            pixelPtr,                                                     // dest
            _videoSurfaceWidthZoomedBytes,       // dest size
            startOfScanlinePtr,                                           // source
            _videoSurfaceWidthZoomedBytes        // source size
        );
    }
}

void _video_copy_visible_paper_screen_line_for_zoom(int y) {
    Uint32* startPtr = _paperScanlineStartPointers[y - HIDE_TOP_BORDER_COUNT];
    // save it for later
    Uint32* startOfScanlinePtr = startPtr;

    // copy paper screen line as many times as zoom
    for (int dup = 1; dup < _zoom; dup++) {
        // move a line down on screen
        startPtr += _videoSurfacePixelWidth;
        memcpy_s(
            startPtr,                                                     // dest
            _zoomedPaperSizeRGBA,  // dest size
            startOfScanlinePtr,                                           // source
            _zoomedPaperSizeRGBA   // source size
        );
    }
}

void _video_write_paper_scan_line(int y, int rowStartAddress) {
    // ============ LEFT BORDER ===============
    if (!_videoIsBorderSolid) {
        // this happens rarely, only when the border has multiple colours within a single
        // frame, which happens only when border changes, or during some static-screen (game menu?)
        // animation on border
        _video_write_left_border_line(y);
    }

    // ============ PAPER ===============
    Uint32* pixelPtr = _paperScanlineStartPointers[y - HIDE_TOP_BORDER_COUNT];
    // save it so we can duplicate it down as many times as zoom
    Uint32* startOfPaperPixelPtr = pixelPtr;

    // a single horizontal 1-pixel thick, 8-pixel wide line
    for (int i = 0; i < SPECTRUM_PAPER_WIDTH; i += 8) {
        // i iterates through each 8-pixel wide line

        Uint8 pixelByte = memory_read8(rowStartAddress + i/8);
        ula_bus_write8(pixelByte);

        int attributeAddress = _attributeAreaAddresses[y - SPECTRUM_BORDER_LINES_BEFORE_PAPER][i/8];
        Uint8 attributeByte = memory_read8(attributeAddress);

        // figure out R, G, B for this 8-pixel wide line
        // attributes
        // bits:       7        6     5   4   3      2   1   0
        //         FLASH   BRIGHT    P2  P1  P0     I2  I1  I0
        Uint8 isBright = (attributeByte & 0x40) >> 6;
        Uint8 isInverse = (attributeByte & 0x80) && _isFlashInverse;

        Uint8 inkColourPaletteEntry = attributeByte & 0x07;
        Uint8 paperColourPaletteEntry = (attributeByte >> 3) & 0x07;

        Uint32* renderedRGBAPtr = _zoomedRGBAByteRenderedPixels[pixelByte][inkColourPaletteEntry][paperColourPaletteEntry][isInverse][isBright];
        if (renderedRGBAPtr == NULL) {
            renderedRGBAPtr = _video_get_rendered_data_byte(pixelByte, inkColourPaletteEntry, paperColourPaletteEntry, isInverse, isBright);
        }

        // render the 8-pixel ZX Spectrum line to a zoomed screen line
        memcpy_s(
            pixelPtr,                                    // dest
            8 * _zoom * sizeof(Uint32),                  // dest size
            renderedRGBAPtr,                             // source
            8 * _zoom * sizeof(Uint32)                   // source size
        );

        pixelPtr += 8 * _zoom;
    }

    // ============ RIGHT BORDER ===============
    if (!_videoIsBorderSolid) {
        // this happens rarely, only when the border has multiple colours within a single
        // frame, which happens only when border changes, or during some static-screen (game menu?)
        // animation on border
        _video_write_right_border_line(y);
    }

    // ============ VERTICAL ZOOM ===============
    // by this point, we have:
    //     1. drawn one screen line of a ZX Spectrum scan line, and
    //     2. either:
    //        a. drew one screen line of left and right borders, or
    //        b. skipped drawing border altogether, because it is drawn in bulk (4 large 
    //           rectangles for the whole screen)

    // we now have to duplicate screen lines down, as many times as zoom (minus one)
    if (!_videoIsBorderSolid) {
        // case [1, 2a]
        // we duplicate entire screen lines
        _video_copy_visible_screen_line_for_zoom(y);
    }
    else {
        // case [1, 2b]
        // we duplicate just paper lines
        _video_copy_visible_paper_screen_line_for_zoom(y);
    }
}

inline Uint8 _video_get_must_render_pixels() {
    if (timing_is_faster()) {
        // turbo mode is on, so rarely draw pixels ...
        if ((_frames & 0x02) == 0x02) {
            // ... but when we do, it must be guaranteed to take place
            _forceFrameRender = 1;
            return 1;
        }

        return 0;
    }
    
    // due to how "is lagging" is computed, without this the frameskip will be infinite
    // when breakpointing during debugging
    // 
    // EXPERIMENTALLY DETERMINED
    Uint64 maxConsecutiveSkippedFrames = 4;
    if (_frameskipIsEnabled && (_consecutiveSkippedFrames >= maxConsecutiveSkippedFrames)) {
        // we're not skipping this frame, because we're at the limit of
        // allowed consecutive skipped frames
        _consecutiveSkippedFrames = 0;
        return 1;
    }

    if (_frameskipIsEnabled && _video_is_lagging()) {
        return 0;
    }

    return 1;
}

void _video_run_CPU_for_frame() {
    int y = 0;

    // ==============================================
    // BORDER BEFORE (ABOVE) PAPER
    // ==============================================
    for (int i = 0; i < SPECTRUM_BORDER_LINES_BEFORE_PAPER; i++) {
        _video_run_cpu_for_scanline();
        _borderColourPerScanline[y] = io_read8_border_colour();
        y++;
    }

    // ==============================================
    // TOP THIRD
    // ==============================================
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        _video_run_cpu_for_scanline();
        _borderColourPerScanline[y] = io_read8_border_colour();
        y++;
    }

    // ==============================================
    // MIDDLE THIRD
    // ==============================================
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        _video_run_cpu_for_scanline();
        _borderColourPerScanline[y] = io_read8_border_colour();
        y++;
    }


    // ==============================================
    // BOTTOM THIRD
    // ==============================================
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        _video_run_cpu_for_scanline();
        _borderColourPerScanline[y] = io_read8_border_colour();
        y++;
    }

    ula_bus_go_inactive();

    // ==============================================
    // BORDER AFTER (BELOW) PAPER
    // ==============================================
    for (int i = 0; i < SPECTRUM_BORDER_LINES_AFTER_PAPER; i++) {
        _video_run_cpu_for_scanline();
        _borderColourPerScanline[y] = io_read8_border_colour();
        y++;
    }
}

void _video__preframe() {
    if (_isDestroying || _timerCallbackDestroyed) {
        _timerCallbackDestroyed = 1;
        return;
    }

    if (_stateLoadRequested) {
        savestates_load(_stateLoadSlot);
        _stateLoadRequested = 0;
    }

    if (_frames == 0) {
        // first frame is guaranteed to be at normal speed, so we 
        // can initialize this timer here
        ftime(&_videoStartSinceNormalSpeed);
    }

    ftime(&_currentFrameStart);

    // reliance on whether it's lagging at all lowers resyncs by a lot
    // interestingly, even though dynamic takes into account actual elapsed time,
    // using only dynamic will resync more than a combination of static/dynamic on 
    // slower computers (approx 20% frameskip)
    // 
    // EXPERIMENTALLY DETERMINED
    double dynamicSoundFrameskipPercentageThreshold = 13.0f;

    switch (_videoSoundSwitch)
    {
    case VideoSoundSwitch_OnLag:
        if (_video_is_lagging()) {
            sound_set_sampling_mode(SoundDynamicSampling);
        }
        else {
            sound_set_sampling_mode(SoundStaticSampling);
        }
        break;
    case VideoSoundSwitch_OnFrameskip:
        if (_video_get_frameskip_percentage() > dynamicSoundFrameskipPercentageThreshold) {
            sound_set_sampling_mode(SoundDynamicSampling);
        }
        else {
            sound_set_sampling_mode(SoundStaticSampling);
        }
        break;
    default:
        break;
    }

    _video_handle_flash();

    memory_mark_frame_start();
    io_mark_frame_start();
    cpu_request_interrupt();

    _video_run_CPU_for_frame();

    // figure out if border is solid (a single colour), since:
    //     - this case is very common, and
    //     - it allows significant border drawing optimization
    _videoIsBorderSolid = 1;
    Uint8 firstScanlineBorderColour = _borderColourPerScanline[0];
    for (int i = 1; i < BORDER_SCANLINES; i++) {
        if (_borderColourPerScanline[i] != firstScanlineBorderColour) {
            _videoIsBorderSolid = 0;
            break;
        }
    }
}

void _video__postframe() {
    if (!timing_is_faster()) {
        _framesAtNormalSpeed++;
        _framesAtNormalSpeedSinceNormalSpeed++;
        
        if (!_lastFrameWasNormalSpeed) {
            // we have switched to normal speed during THIS frame
            _framesAtNormalSpeedSinceNormalSpeed = 0;
            ftime(&_videoStartSinceNormalSpeed);
        }
        _lastFrameWasNormalSpeed = 1;
    }
    else {
        _lastFrameWasNormalSpeed = 0;
    }

    double msPerFrame = (double)timing_get_target_milliseconds_per_video_frame();
    Uint64 framesPerSecond = (Uint64)(1000.0f / msPerFrame);
    Uint64 soundFrameDelay = framesPerSecond / 2;
    if (_frames == 0) {
        // delay start of sound to accumulate some samples
        sound_play();
    }

    _frames++;

    double averageMsPerFrame = _video_get_actual_average_milliseconds_per_frame();
    timing_set_actual_milliseconds_per_video_frame(averageMsPerFrame);
    
    sound_notify_end_of_frame();

    // compute this frame's duration
    struct timeb currentFrameEnd;
    ftime(&currentFrameEnd);
    Uint64 frameTimeMs = (Uint64)(1000.0 * (currentFrameEnd.time - _currentFrameStart.time)
        + (currentFrameEnd.millitm - _currentFrameStart.millitm));

    // record this frame's excess time as idle time if this frame took an extremely long time
    // this is to account for:
    //     - poor scheduling of this thread by OS, or
    //     - debugging and breakpointing, causing video frames to take an extreme amount of time

    // EXPERIMENTALLY DETERMINED
    Uint64 idleFrameTimeLimit = 10 * (Uint64)timing_get_target_milliseconds_per_video_frame();
    if (frameTimeMs > idleFrameTimeLimit) {
        // this frame includes idle time
        Uint64 idleTime = frameTimeMs - timing_get_target_milliseconds_per_video_frame();
        timing_record_idle_time(idleTime);
    }

    if (_stateSaveRequested) {
        savestates_save(_stateSaveSlot);
        _stateSaveRequested = 0;
    }
}

Uint8 _video__uis(Uint8 forceRenderUis) {
    // render other modules, but only if something interesting happened
    Uint8 renderAnyUis = forceRenderUis;

    if (tapui_render(_renderer, forceRenderUis)) {
        renderAnyUis = 1;
    }

    if (pokeui_render(_renderer, forceRenderUis)) {
        renderAnyUis = 1;
    }

    if (notification_render(_renderer)) {
        renderAnyUis = 1;
    }

    return renderAnyUis;
}

void _video__DEBUG_draw_ZX_Spectrum_screen_outline() {
    SDL_SetRenderDrawColor(_renderer, 255, 0, 0, 255);
    SDL_Rect _videoSpectrumVisible;
    _videoSpectrumVisible.x = 0;
    _videoSpectrumVisible.y = 0;
    _videoSpectrumVisible.w = VISIBLE_SPECTRUM_SCREEN_WIDTH * _zoom;
    _videoSpectrumVisible.h = VISIBLE_SPECTRUM_SCREEN_HEIGHT * _zoom;
    SDL_RenderFillRect(_renderer, &_videoSpectrumVisible);

    SDL_SetRenderDrawColor(_renderer, 0, 0, 255, 255);
    _videoSpectrumVisible.x++;
    _videoSpectrumVisible.y++;
    _videoSpectrumVisible.w-=2;
    _videoSpectrumVisible.h-=2;
    SDL_RenderFillRect(_renderer, &_videoSpectrumVisible);
}

// draws the pixels which are part of the spectrum screen (as in, no UIs, etc.)
//
void _video__spectrum_pixels(Uint8 isHardwareRenderer) {
    SDL_LockSurface(_surface);

    // ==============================================
    // BORDER BEFORE (ABOVE) PAPER
    // ==============================================
    int y = HIDE_TOP_BORDER_COUNT;

    if (!_videoIsBorderSolid) {
        for (int i = 0; i < VISIBLE_SPECTRUM_TOP_BORDER_COUNT; i++) {
            _video_write_border_scan_line(y + i);
        }
    }
    y += VISIBLE_SPECTRUM_TOP_BORDER_COUNT;


    // ==============================================
    // TOP THIRD
    // ==============================================
    int rowStartAddressAtStartOfThird = SPECTRUM_VIDEO_RAM_START;

    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        int mod = k % 8;
        int div = k / 8;
        int rowStartAddress = rowStartAddressAtStartOfThird + (mod * SPECTRUM_PAPER_HEIGHT / 24) * (SPECTRUM_PAPER_WIDTH / 8) + (div * SPECTRUM_PAPER_WIDTH / 8);
        _video_write_paper_scan_line(y + k, rowStartAddress);
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;
    rowStartAddressAtStartOfThird += (SPECTRUM_PAPER_WIDTH / 8) * (SPECTRUM_PAPER_HEIGHT / 3);


    // ==============================================
    // MIDDLE THIRD
    // ==============================================
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        int mod = k % 8;
        int div = k / 8;
        int rowStartAddress = rowStartAddressAtStartOfThird + (mod * SPECTRUM_PAPER_HEIGHT / 24) * (SPECTRUM_PAPER_WIDTH / 8) + (div * SPECTRUM_PAPER_WIDTH / 8);
        _video_write_paper_scan_line(y + k, rowStartAddress);
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;
    rowStartAddressAtStartOfThird += (SPECTRUM_PAPER_WIDTH / 8) * (SPECTRUM_PAPER_HEIGHT / 3);


    // ==============================================
    // BOTTOM THIRD
    // ==============================================
    for (int k = 0; k < SPECTRUM_PAPER_HEIGHT / 3; k++) {
        int mod = k % 8;
        int div = k / 8;
        int rowStartAddress = rowStartAddressAtStartOfThird + (mod * SPECTRUM_PAPER_HEIGHT / 24) * (SPECTRUM_PAPER_WIDTH / 8) + (div * SPECTRUM_PAPER_WIDTH / 8);
        _video_write_paper_scan_line(y + k, rowStartAddress);
    }
    y += SPECTRUM_PAPER_HEIGHT / 3;


    // ==============================================
    // BORDER AFTER (BELOW) PAPER
    // ==============================================
    if (!_videoIsBorderSolid) {
        for (int i = 0; i < SPECTRUM_BORDER_LINES_AFTER_PAPER - HIDE_BOTTOM_BORDER_COUNT; i++) {
            _video_write_border_scan_line(y + i);
        }
    }

    SDL_UnlockSurface(_surface);

    // this is unnecessary when using a software renderer
    if (isHardwareRenderer) {
        // convert to a texture the surface to which we've drawn the pixels
        SDL_Texture* entireWindowTexture = SDL_CreateTextureFromSurface(_renderer, _surface);

        // render play area
        SDL_RenderCopy(_renderer, entireWindowTexture, NULL, &_entireWindowRectangle);
        SDL_DestroyTexture(entireWindowTexture);
    }

    // ===================================================
    // BORDER OPTIMIZATION
    //     if border is a single colour, then draw
    //     it more efficiently than scan line by scan line
    // ===================================================
    if (_videoIsBorderSolid) {
        // since all scan lines have the same border colour, use the colour of the first scan line
        Uint8 colourOffset = 3 * _borderColourPerScanline[0];
        Uint8 red = *(_palette + colourOffset + 0);
        Uint8 green = *(_palette + colourOffset + 1);
        Uint8 blue = *(_palette + colourOffset + 2);
        SDL_SetRenderDrawColor(_renderer, red, green, blue, 255);

        SDL_RenderFillRect(_renderer, &_topBorder);
        SDL_RenderFillRect(_renderer, &_bottomBorder);
        SDL_RenderFillRect(_renderer, &_leftBorder);
        SDL_RenderFillRect(_renderer, &_rightBorder);
    }

    //_video__DEBUG_draw_ZX_Spectrum_screen_outline();
}

void _video__spectrum_pixels__wrapper(Uint8 isHardwareRenderer) {
    DWORD dwWaitResult = WaitForSingleObject(_pixelCacheMutex, INFINITE);

    switch (dwWaitResult) {
    case WAIT_OBJECT_0:
        // this thread got ownership of the mutex
        __try {
            _video__spectrum_pixels(isHardwareRenderer);
        }

        __finally {
            if (!ReleaseMutex(_pixelCacheMutex))
            {
                // an error happened
            }
        }
        break;

    case WAIT_ABANDONED:
        // this thread got ownership of an abandoned mutex
        return;
    }
}

void _video__frame_pixels(enum frameRenderMode mode) {
    Uint8 isHardwareRenderer = _rendererFlags != SDL_RENDERER_SOFTWARE;
    Uint8 drawPixels = mode != ForceSkipPixels;

    // render ZX Spectrum screen pixels
    if (mode != ForceSkipPixels) {
        _video__spectrum_pixels__wrapper(isHardwareRenderer);
    }
    
    // render any other pixels (such as for various UIs)
    Uint8 forceUiRendering = isHardwareRenderer;
    Uint8 uisRendered = _video__uis(forceUiRendering);

    if (uisRendered || drawPixels) {
        SDL_RenderPresent(_renderer);
    }

    if (drawPixels) {
        _forceFrameRender = 0;
    }
}

// Top level timer-independent (that is, called by all different runners: SDL, QueueTimer, etc.)
// per-frame-invoked function
void _video_frame_runner() {
    _video__preframe();

    Uint8 mustRenderPixels = _video_get_must_render_pixels();
    Uint8 videoHasChanged = memory_was_video_memory_modified_this_frame() || io_is_border_modified() || !tapui_is_tape_stopped() || _forceFrameRender;

    if (mustRenderPixels) {
        if (videoHasChanged) {
            _video__frame_pixels(DrawPixels);
        }
        else {
            _framesWithNoVideoChanges++;
            _consecutiveSkippedFrames++;
            _video__frame_pixels(ForceSkipPixels);
        }
    }
    else {
        _consecutiveSkippedFrames++;
        if (!timing_is_faster()) {
            _framesSkipped++;
        }
        _video__frame_pixels(ForceSkipPixels);
    }
    _video__postframe();
}

Uint32 _video_callback__via_SDL_Timer(void* param) {
    struct timeb now;
    ftime(&now);

    Uint64 runTimeMs = (Uint64)(1000.0 * (now.time - _videoStart.time)
        + (now.millitm - _videoStart.millitm));
    if (runTimeMs > (Uint64)VIDEO_DELAY_TIMER_START_MS) {
        // we do nothing for the first bit of time
        _video_frame_runner();
    }

    return timing_get_target_milliseconds_per_video_frame();
}

void __stdcall _video_callback__via_QueueTimer(PVOID lpParameter, BOOLEAN TimerOrWaitFired) {
    DWORD dwWaitResult = WaitForSingleObject(_videoFrameMutex, INFINITE);
    
    switch (dwWaitResult) {
        case WAIT_OBJECT_0:
            // this thread got ownership of the mutex
            __try {
                _video_frame_runner();
            }

            __finally {
                if (!ReleaseMutex(_videoFrameMutex))
                {
                    // an error happened
                }
            }
            break;

        case WAIT_ABANDONED:
            // this thread got ownership of an abandoned mutex
            return;
    }
}

void video_disable_frameskip() {
    _frameskipIsEnabled = 0;
}

void _video_set_zoom(Uint8 zoom) {
    _zoom = zoom;
}

BOOL video_start__via_SDL_Timer(SDL_Window* window, int windowWidth, int windowHeight) {
    log_write("Using a SDL Timer-based video callback");
    
    _frames = 0;
    _framesAtNormalSpeed = 0;
    _framesAtNormalSpeedSinceNormalSpeed = 0;
    _lastFrameWasNormalSpeed = !timing_is_faster();
    _tstates = 0;

    _isDestroying = 0;
    _timerCallbackDestroyed = 0;
    ftime(&_videoStart);

    _entireWindowRectangle.x = 0;
    _entireWindowRectangle.y = 0;
    _entireWindowRectangle.w = _surface->w;
    _entireWindowRectangle.h = _surface->h;

    // hardware renderers crash when the screen locks, resulting in
    // frozen video
    _renderer = SDL_CreateRenderer(window, -1, _rendererFlags);
    if (_renderer == NULL) {
        log_write("Error: could not get a video renderer");
        return FALSE;
    }
    _video_log_renderer_choice();

    _videoIsInitialized = TRUE;

    _timer = SDL_AddTimer(timing_get_target_milliseconds_per_video_frame(), (SDL_TimerCallback)_video_callback__via_SDL_Timer, NULL);
    log_write("Started timer, yielding CPU control to it");
    return TRUE;
}

BOOL video_start__via_QueueTimer(SDL_Window* window, int windowWidth, int windowHeight) {
    log_write("Using a QueueTimer-based video callback");
    _frames = 0;
    _framesAtNormalSpeed = 0;
    _framesAtNormalSpeedSinceNormalSpeed = 0;
    _lastFrameWasNormalSpeed = !timing_is_faster();
    _tstates = 0;

    _isDestroying = 0;
    _timerCallbackDestroyed = 0;
    ftime(&_videoStart);

    _entireWindowRectangle.x = 0;
    _entireWindowRectangle.y = 0;
    _entireWindowRectangle.w = _surface->w;
    _entireWindowRectangle.h = _surface->h;

    // hardware renderers crash when the screen locks, resulting in
    // frozen video
    _renderer = SDL_CreateRenderer(window, -1, _rendererFlags);
    if (_renderer == NULL) {
        log_write("Error: could not get a video renderer");
        return FALSE;
    }
    _video_log_renderer_choice();
    
    _videoFrameMutex = CreateMutex(NULL, FALSE, NULL);
    if (!_videoFrameMutex) {
        log_write("Error: unable to create video frame timer mutex");
        return FALSE;
    }
    
    Uint16 framePeriod = timing_get_target_milliseconds_per_video_frame();
    ULONG flags = WT_EXECUTEINTIMERTHREAD;
    BOOL res = CreateTimerQueueTimer(&_callbackTimerHandle, NULL, _video_callback__via_QueueTimer, NULL, VIDEO_DELAY_TIMER_START_MS, framePeriod, flags);
    if (!res) {
        log_write("Error: unable to start video frame timer");
        _callbackTimerHandle = NULL;
        return FALSE;
    }
    
    _videoIsInitialized = TRUE;
    log_write("Started video timer, yielding CPU control to it");
    return TRUE;
}

SDL_Renderer* video_get_renderer() {
    return _renderer;
}

void _video_report_statistics() {
    char* message = (char*)malloc(500 * sizeof(char));
    if (message == NULL) {
        return;
    }

    Uint64 timeSpentWhileCPUFasterMs = timing_get_total_real_time_spent_while_faster();
    Uint64 totalTimeMs = _video_get_runtime_ms();
    Uint64 runTimeMs = totalTimeMs - timeSpentWhileCPUFasterMs;

    sprintf_s(message, 490, "Zoom: %u", _zoom);
    log_write(message);

    if (_frameskipIsEnabled) {
        log_write("Frameskip is enabled");
        sprintf_s(message, 490, "Lag allowance before frameskip: %u%%", (Uint16)round(_lagAllowanceBeforeFrameskip * 100.0f));
        log_write(message);
    }
    else {
        log_write("Frameskip is disabled");
    }

    switch (_videoSoundSwitch)
    {
    case VideoSoundSwitch_OnLag:
        log_write("Video sound switching: 0 (on lag)");
        break;
    case VideoSoundSwitch_OnFrameskip:
        log_write("Video sound switching: 1 (on frameskip)");
        break;
    default:
        break;
    }

    long long totalPrerenderedDataEntries = 256 * SPECTRUM_COLOUR_COUNT * SPECTRUM_COLOUR_COUNT * 2 * 2;
    double cachedPixelBytesPercentage = (double)_prerenderedDataByteCount / (double)totalPrerenderedDataEntries * 100.0f;

    long long cachedPixelSizeBytes = _prerenderedDataByteCount * 8 * _zoom * sizeof(Uint32);
    long long cachedPixelSizeKilobytes = cachedPixelSizeBytes / 1024;

    if (cachedPixelSizeBytes % 1024 != 0) {
        cachedPixelSizeKilobytes++;
    }

    if (cachedPixelSizeKilobytes == 0) {
        cachedPixelSizeKilobytes = 1;
    }

    sprintf_s(message, 490, "Pixel cache usage: %I64d entries (%.1f%% of available), %I64dkb", 
        _prerenderedDataByteCount, 
        cachedPixelBytesPercentage, 
        cachedPixelSizeKilobytes);
    log_write(message);

    double frames = (double)_framesAtNormalSpeed;
    sprintf_s(message, 490, "Total video frames (rendered and skipped): %lld", (Uint64)frames);
    log_write(message);

    double percentage = _video_get_frameskip_percentage();
    sprintf_s(message, 490, "Frames skipped: %I64d (%.1f%% of total)", _framesSkipped, percentage);
    log_write(message);

    double framesRepeatedPercentage = _video_get_framesrepeated_percentage();
    sprintf_s(message, 490, "Frames repeated (no visible changes): %I64d (%.1f%% of total)", _framesWithNoVideoChanges, framesRepeatedPercentage);
    log_write(message);

    double fps = frames / ((double)runTimeMs / 1000.f);
    sprintf_s(message, 490, "Average frames per second: %.2f", fps);
    log_write(message);

    double millisPerFrame = (double)runTimeMs / frames;
    Uint32 targetMillisPerFrame = timing_get_target_milliseconds_per_video_frame();
    sprintf_s(message, 490, "Average milliseconds per frame: %.3f (target is %u)", millisPerFrame, targetMillisPerFrame);
    log_write(message);

    long double tstates = (long double)_tstates;
    long double tstatesPerVideoFrame = tstates / frames;
    Uint64 targetTstatesPerVideoFrame = timing_get_states_per_video_frame();
    percentage = (tstatesPerVideoFrame / (long double)targetTstatesPerVideoFrame) * 100.0f;
    sprintf_s(message, 490, "Average tstates per video frame: %.2f (target is %lld), %.1f%%", 
        tstatesPerVideoFrame, 
        targetTstatesPerVideoFrame,
        percentage);
    log_write(message);

    log_write("\nDURATION");
    sprintf_s(message, 490, "Total real duration (CPU at normal speed): %I64u ms", runTimeMs);
    log_write(message);
    sprintf_s(message, 490, "Total real duration (CPU at faster speed): %I64u ms", timeSpentWhileCPUFasterMs);
    log_write(message);
    sprintf_s(message, 490, "Total real duration (CPU normal and fast): %I64u ms", totalTimeMs);
    log_write(message);

    free(message);
}

void video_destroy__via_SDL_Timer() {
    log_write("Destroying SDL_Timer-based video");
    _isDestroying = 1;
    while (!_timerCallbackDestroyed) {}

    SDL_RemoveTimer(_timer);
    ftime(&_videoEnd);
}

void video_destroy__via_QueueTimer() {
    log_write("Destroying QueueTimer-based video");
    if (_callbackTimerHandle) {
        _isDestroying = 1;
        // we busy-wait for callback to become NOOP only if we actually initialized
        // the callback timer successfully
        while (!_timerCallbackDestroyed) {}

        if (!DeleteTimerQueueTimer(NULL, _callbackTimerHandle, NULL)) {
            log_write("Warning: deletion of QueueTimer failed");
        }
    }

    ftime(&_videoEnd);

    if (_videoFrameMutex) {
        CloseHandle(_videoFrameMutex);
    }

    if (_surface) {
        SDL_FreeSurface(_surface);
    }
}

BOOL video_start(enum videoMode videoMode, SDL_Window* window, char zoom, int windowWidth, int windowHeight) {
    BOOL result;

    _videoMode = videoMode;
    _video_set_zoom(zoom);

    _topBorder.x = 0;
    _topBorder.y = 0;
    _topBorder.w = VISIBLE_SPECTRUM_SCREEN_WIDTH * _zoom;
    _topBorder.h = VISIBLE_SPECTRUM_TOP_BORDER_COUNT * _zoom;

    _bottomBorder.x = 0;
    _bottomBorder.y = (VISIBLE_SPECTRUM_TOP_BORDER_COUNT + SPECTRUM_PAPER_HEIGHT) * _zoom;
    _bottomBorder.w = VISIBLE_SPECTRUM_SCREEN_WIDTH * _zoom;
    _bottomBorder.h = VISIBLE_SPECTRUM_BOTTOM_BORDER_COUNT * _zoom;

    _leftBorder.x = 0;
    _leftBorder.y = VISIBLE_SPECTRUM_TOP_BORDER_COUNT * _zoom;
    _leftBorder.w = VISIBLE_SPECTRUM_LEFT_BORDER_WIDTH * _zoom;
    _leftBorder.h = SPECTRUM_PAPER_HEIGHT * _zoom;

    _rightBorder.x = (VISIBLE_SPECTRUM_SCREEN_WIDTH - VISIBLE_SPECTRUM_RIGHT_BORDER_WIDTH) * _zoom;
    _rightBorder.y = VISIBLE_SPECTRUM_TOP_BORDER_COUNT * _zoom;
    _rightBorder.w = VISIBLE_SPECTRUM_RIGHT_BORDER_WIDTH * _zoom;
    _rightBorder.h = SPECTRUM_PAPER_HEIGHT * _zoom;

    _window = window;
    _surface = SDL_GetWindowSurface(window);
    if (!_surface) {
        log_write("Error: unable to create SDL window surface");
        return FALSE;
    }
    _videoSurfacePixelWidth = _surface->pitch / sizeof(Uint32);
    _videoSurfaceWidthZoomedBytes = _surface->pitch;

    _video_compute_rgba_palette(_surface);
    _video_compute_scanline_pointers(_surface);
    _video_compute_visible_screen_scanline_pointers(_surface);
    _video_compute_attribute_addresses(_surface);

    _zoomedPaperSizeRGBA = SPECTRUM_PAPER_WIDTH * _zoom * sizeof(Uint32);
    _zoomedVisibleSpectrumScanLineSizeRGBA = VISIBLE_SPECTRUM_SCREEN_WIDTH * _zoom * sizeof(Uint32);

    _pixelCacheMutex = CreateMutex(NULL, FALSE, NULL);
    if (!_pixelCacheMutex) {
        log_write("Error: unable to create pixel cache mutex");
        return FALSE;
    }

    switch (_videoMode)
    {
    case Video_viaQueueTimer:
        result = video_start__via_QueueTimer(window, windowWidth, windowHeight);
        return result;
        break;
    case Video_viaSDLTimer:
        result = video_start__via_SDL_Timer(window, windowWidth, windowHeight);
        return result;
        break;
    default:
        break;
    }

    return FALSE;
}

void video_destroy() {
    log_write("\nVIDEO");

    switch (_videoMode)
    {
    case Video_viaQueueTimer:
        video_destroy__via_QueueTimer();
        break;
    case Video_viaSDLTimer:
        video_destroy__via_SDL_Timer();
        break;
    default:
        break;
    }

    _video_destroy_prerendered_data_bytes__wrapper();
    _video_report_statistics();

    if (_pixelCacheMutex) {
        CloseHandle(_pixelCacheMutex);
    }
}

void video_set_soundswitch(enum videoSoundSwitch soundSwitch) {
    _videoSoundSwitch = soundSwitch;
}

void video_state_save(Uint8 slot) {
    _stateSaveSlot = slot;
    MemoryBarrier();
    _stateSaveRequested = 1;
}

void video_state_load(Uint8 slot) {
    _stateLoadSlot = slot;
    MemoryBarrier();
    _stateLoadRequested = 1;
}

void video_set_renderer_type(Uint8 renderer) {
    switch (renderer) {
    case 0:
        _rendererFlags = SDL_RENDERER_SOFTWARE;
        break;
    case 1:
        _rendererFlags = SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC;
        break;
    default:
        _rendererFlags = SDL_RENDERER_SOFTWARE;
        break;
    }
}